import time
import urllib.request
from typing import Callable
from datetime import datetime
from ssl import SSLCertVerificationError
from urllib.error import URLError, HTTPError

import ipih
from CheckerService.const import *
from pih import A, PIHThreadPoolExecutor
from pih.tools import nn, ne, e, j, js, lw, one
from pih.collections import (
    User,
    Result,
    ZabbixMetrics,
    ResourceStatus,
    DiskStatistics,
    WSResourceStatus,
    SiteResourceStatus,
    DisksStatisticsStatus,
    ComputerDescription,
    ServerResourceStatus,
)


class CheckerApi:

    def __init__(self) -> None:

        self.resources_status_map: dict[str, ResourceStatus] = {}
        self.ws_status_map: dict[str, WSResourceStatus] = {}
        self.server_status_map: dict[str, ServerResourceStatus] = {}
        self.disk_resource_status_map: dict[str, DisksStatisticsStatus] = {}
        self.resources_accessable_map: dict[str, bool] = {}
        self.check_counter: int = 0

        for wappi_profile in A.CT_ME_WH_W.Profiles:
            RESOURCE_LIST.WAPPI.append(
                ResourceStatus(
                    wappi_profile.value,  # type: ignore
                    js(("Профиль Wappi:", wappi_profile.name)),
                    (2, 50, 24 * 60),
                )
            )

        for site_resource_description in A.CT_R_D.SITE_LIST:
            RESOURCE_LIST.SITE.append(
                A.D.fill_data_from_source(
                    SiteResourceStatus(), site_resource_description
                )  # type: ignore
            )

        def filter_function(workstation: ComputerDescription) -> bool:
            return A.C_WS.watchable(workstation)

        def ws_every_action(workstation: ComputerDescription) -> None:
            RESOURCE_LIST.WS.append(
                WSResourceStatus(
                    workstation.name,
                    j(
                        (
                            "Компьютер: ",
                            workstation.name,
                            " (",
                            workstation.description,
                            ")",
                        )
                    ),
                )
            )

        ws_description_result_list: Result[list[ComputerDescription]] = (
            A.R_WS.all_description()
        )
        A.R.every(
            ws_every_action,
            A.R.filter(filter_function, ws_description_result_list, True),
        )

        def server_every_action(server: ComputerDescription) -> None:
            RESOURCE_LIST.SERVER.append(
                ServerResourceStatus(
                    server.name,
                    j(("Сервер: ", server.name, " (", server.description, ")")),
                )
            )

        server_description_result_list: Result[list[ComputerDescription]] = (
            A.R_SRVS.all_description()
        )
        A.R.every(server_every_action, server_description_result_list)

        computer_description_list: list[ComputerDescription] = (
            ws_description_result_list.data + server_description_result_list.data
        )  # type: ignore

        for index in (0, 1):
            RESOURCE_LIST.DISK_STATISTICS_COMPUTER_DESCRIPTION.extend(
                A.D.map(
                    lambda item: (item, index),
                    A.D_FL.computers_with_property(
                        computer_description_list,
                        (
                            A.CT_AD.ComputerProperties.DiskReportable
                            if index == 0
                            else A.CT_AD.ComputerProperties.DiskReportableViaZabbix
                        ),
                    ),
                )
            )

    def send_message_to_it_users(self, message: str) -> None:
        def every_action(user: User) -> None:
            A.ME_WS.to_user(user, message)

        A.R.every(every_action, A.R_U.by_job_position(A.CT_AD.JobPositions.IT))

    def update_resource_status_and_notify(
        self,
        value: ResourceStatus,
        accessable: bool,
        on_init_handler: Callable[[ResourceStatus, bool], None] | None = None,
        on_access_handler: Callable[[ResourceStatus, bool], None] | None = None,
        on_inaccess_handler: Callable[[ResourceStatus, bool], None] | None = None,
    ) -> bool:
        self.resources_accessable_map[value.address] = value.accessable  # type: ignore
        if (
            accessable
            and isinstance(value, SiteResourceStatus)
            and value.check_certificate_status
            and ne(value.certificate_status)
        ):
            accessable = lw(value.certificate_status).find("invalid") == -1
            if accessable:
                pass
        if accessable:
            value.inaccessibility_counter = 0
        else:
            value.inaccessibility_counter += 1
            value.inaccessibility_counter_total += 1
        accessable = A.C_R.accessibility(value)  # type: ignore
        value.accessable = accessable
        if isinstance(value, WSResourceStatus):
            self.ws_status_map[value.address] = value  # type: ignore
        elif isinstance(value, ServerResourceStatus):
            self.server_status_map[value.address] = value  # type: ignore
        else:
            self.resources_status_map[value.address] = value  # type: ignore

        if self.check_counter == value.inaccessibility_check_values[0]:
            if nn(on_init_handler):
                on_init_handler(value, accessable)  # type: ignore
        else:
            if accessable:
                if self.resources_accessable_map[value.address] == False and nn(  # type: ignore
                    on_access_handler
                ):
                    on_access_handler(value, value.inaccessibility_counter == 0)  # type: ignore
            else:
                if (
                    value.inaccessibility_counter
                    <= value.inaccessibility_check_values[0]
                    + value.inaccessibility_check_values[1]
                ) or (
                    value.inaccessibility_counter
                    % value.inaccessibility_check_values[2]
                    == 0
                ):
                    if nn(on_inaccess_handler):
                        on_inaccess_handler(value, self.at_first_time(value))  # type: ignore
        return accessable

    def on_inaccess_handler(
        self, resource_status: ResourceStatus, at_first_time: bool
    ) -> None:
        A.E.resource_inaccessible(resource_status, at_first_time)

    def on_access_handler(
        self, resource_status: ResourceStatus, at_first_time: bool
    ) -> None:
        # if at_first_time:
        A.E.resource_accessible(resource_status, at_first_time)

    def on_init_handler(
        self, resource_status: ResourceStatus, accessable: bool
    ) -> None:
        (
            A.E.resource_accessible(resource_status, True)
            if accessable
            else A.E.resource_inaccessible(resource_status, True)
        )

    def at_first_time(self, resource_status: ResourceStatus) -> bool:
        return (
            resource_status.inaccessibility_counter
            == resource_status.inaccessibility_check_values[0],
        )  # type: ignore

    def set_disk_resource_status(
        self, value: ComputerDescription, via_zabbix: bool = False
    ) -> None:
        def set_disk_statistics_status_for_host(
            host: str, disk_statistics: DiskStatistics
        ) -> None:
            if disk_statistics.size != 0:
                if host not in self.disk_resource_status_map:
                    self.disk_resource_status_map[host] = DisksStatisticsStatus(host)
                self.disk_resource_status_map[host].disk_list.append(disk_statistics)

        disk_statistics_list: list[DiskStatistics] | None = None
        if via_zabbix:
            disk_statistics_map: dict[str, DiskStatistics] = {}
            host_id: int = one(
                A.R.map(
                    lambda item: item.id,
                    A.R.filter(
                        lambda item: lw(item.name) == lw(value.name), A.R_Z.hosts()
                    ),
                )  # type: ignore
            )  # type: ignore

            item_list: list[ZabbixMetrics] = A.R.filter(
                lambda item: item.key_.find("vfs.fs.size") != -1,  # type: ignore
                A.R_Z.items(host_id),
            ).data  # type: ignore
            for item in item_list:
                key_: str = item.key_  # type: ignore
                part_index: int = key_.find(",")
                name: str = key_[key_.find("[") + 1 : part_index - 1]
                type_value: str = key_[part_index + 1 : key_.find("]")]
                if name not in disk_statistics_map:
                    disk_statistics_map[name] = DiskStatistics(name)
                if type_value == "total":
                    disk_statistics_map[name].size = int(item.lastvalue)
                if type_value == "used":
                    disk_statistics_map[name].free_space = disk_statistics_map[
                        name
                    ].size - int(  # type: ignore
                        item.lastvalue
                    )
            disk_statistics_list = A.D.to_list(disk_statistics_map)
        else:
            disk_statistics_list = A.EXC.get_disk_statistics_list(value.name)  # type: ignore
        A.D.every(
            lambda item: set_disk_statistics_status_for_host(value.name, item),  # type: ignore
            disk_statistics_list,
        )

    def ws_or_server_check(
        self,
        resource_status: ResourceStatus,
        check_function: Callable[[ResourceStatus], bool],
        events: tuple[A.CT_E, ...],
    ) -> None:
        def on_access_handler(
            resource_status: ResourceStatus, at_first_time: bool
        ) -> None:
            if at_first_time:
                A.E.send(events[0], (resource_status.address,))

        def on_inaccess_handler(
            resource_status: ResourceStatus, at_first_time: bool
        ) -> None:
            if at_first_time:
                A.E.send(events[1], (resource_status.address,))

        self.update_resource_status_and_notify(
            resource_status,
            check_function(resource_status),
            lambda _, accessable: (
                on_access_handler(resource_status, True)
                if accessable
                else on_inaccess_handler(resource_status, True)
            ),
            on_access_handler,
            on_inaccess_handler,
        )

    def check_resources(
        self, force_update: bool = False, current_datetime: datetime | None = None
    ) -> None:
        start: float = time.perf_counter()
        self.check_counter += 1
        self.action_at_work = True
        #
        self.update_resource_status_and_notify(
            RESOURCE_LIST.INTERNET,
            A.C_R.accessibility_by_ping(RESOURCE_LIST.INTERNET.address, None, 2, False),  # type: ignore
        )
        site_check_free_space_perion_in_minutes: int = (
            A.S_R.site_check_free_space_perion_in_minutes()
        )
        site_check_certificate_start_time: datetime = (
            A.S_R.site_check_certificate_start_time()
        )
        if not RESOURCE_LIST.INTERNET.accessable:
            self.send_message_to_it_users("Вероятно, интернет отсутствует")
        else:

            def check_site(site_resource_status: SiteResourceStatus) -> None:
                site_address: str = site_resource_status.address  # type: ignore
                try:
                    if force_update or ne(current_datetime):
                        if site_resource_status.check_free_space_status:
                            if (
                                e(site_resource_status.free_space_status)
                                or force_update
                                or (
                                    (
                                        (A.SE.life_time.total_seconds() // 60)
                                        % site_check_free_space_perion_in_minutes
                                    )
                                    == 0
                                )
                            ):
                                site_resource_status.free_space_status = A.R_SSH.get_unix_free_space_information_by_drive_name(
                                    site_resource_status.driver_name, site_address  # type: ignore
                                ).data
                        if site_resource_status.check_certificate_status:
                            if (
                                e(site_resource_status.certificate_status)
                                or force_update
                                or (
                                    ne(current_datetime)
                                    and A.D.is_equal_by_time(
                                        current_datetime,  # type: ignore
                                        site_check_certificate_start_time,
                                    )
                                )
                            ):
                                site_resource_status.certificate_status = (
                                    A.R_SSH.get_certificate_information(
                                        site_address
                                    ).data
                                )
                    url: str = j(
                        (
                            A.D.check(
                                site_resource_status.internal,
                                A.CT.UNTRUST_SITE_PROTOCOL,
                                A.CT.SITE_PROTOCOL,
                            ),
                            site_address,
                        )
                    )
                    self.update_resource_status_and_notify(
                        site_resource_status,
                        urllib.request.urlopen(url).getcode() == 200,
                        self.on_init_handler,
                        self.on_access_handler,
                        self.on_inaccess_handler,
                    )
                except URLError as error:
                    if isinstance(error.reason, SSLCertVerificationError):
                        self.update_resource_status_and_notify(
                            site_resource_status, False
                        )
                        A.E.resource_inaccessible(
                            site_resource_status,
                            self.at_first_time(site_resource_status),
                            A.CT_R_IR.CERTIFICATE_ERROR,
                        )

                    if isinstance(error, HTTPError) and error.code == 503:
                        self.update_resource_status_and_notify(
                            site_resource_status, False
                        )
                        A.E.resource_inaccessible(
                            site_resource_status,
                            self.at_first_time(site_resource_status),
                            A.CT_R_IR.SERVICE_UNAVAILABLE,
                        )

            #
            with PIHThreadPoolExecutor(max_workers=len(RESOURCE_LIST.SITE)) as executor:
                for site_resource_status in RESOURCE_LIST.SITE:
                    executor.submit(check_site, site_resource_status)
            #
            for wappi_resource in RESOURCE_LIST.WAPPI:
                self.update_resource_status_and_notify(
                    wappi_resource,
                    A.C_R.wappi_profile_accessibility(wappi_resource.address),  # type: ignore
                    self.on_init_handler,
                    self.on_access_handler,
                    self.on_inaccess_handler,
                )
            #
            if self.update_resource_status_and_notify(
                RESOURCE_LIST.VPN_PACS_SPB,
                A.C_R.vpn_pacs_accessibility(),
                self.on_init_handler,
                self.on_access_handler,
                self.on_inaccess_handler,
            ):
                #
                self.update_resource_status_and_notify(
                    RESOURCE_LIST.PACS_SPB,
                    A.C_R.pacs_accessibility(),
                    self.on_init_handler,
                    self.on_access_handler,
                    self.on_inaccess_handler,
                )
            else:
                self.update_resource_status_and_notify(RESOURCE_LIST.PACS_SPB, False)
        #
        self.update_resource_status_and_notify(
            RESOURCE_LIST.POLIBASE1,
            A.C_R.polibase_accessibility(),
            self.on_init_handler,
            self.on_access_handler,
            self.on_inaccess_handler,
        )
        self.update_resource_status_and_notify(
            RESOURCE_LIST.POLIBASE2,
            A.C_R.polibase_accessibility(test=True),
            self.on_init_handler,
            self.on_access_handler,
            self.on_inaccess_handler,
        )
        if not RESOURCE_LIST.POLIBASE1.accessable:
            self.send_message_to_it_users("Полибейс недоступен")
        #
        with PIHThreadPoolExecutor(max_workers=len(RESOURCE_LIST.WS)) as executor:
            for ws_resource_status in RESOURCE_LIST.WS:
                executor.submit(
                    self.ws_or_server_check,
                    ws_resource_status,
                    lambda resource_status: A.C_R.accessibility_by_smb_port(
                        resource_status.address, count=2, check_all=False
                    ),
                    (
                        A.CT_E.WATCHABLE_WORKSTATION_IS_ACCESSABLE,
                        A.CT_E.WATCHABLE_WORKSTATION_IS_NOT_ACCESSABLE,
                    ),
                )
        #
        with PIHThreadPoolExecutor(max_workers=len(RESOURCE_LIST.SERVER)) as executor:
            for server_resource_status in RESOURCE_LIST.SERVER:
                executor.submit(
                    self.ws_or_server_check,
                    server_resource_status,
                    lambda resource_status: A.C_R.accessibility_by_ping(
                        resource_status.address, count=2, check_all=False
                    ),
                    (A.CT_E.SERVER_IS_ACCESSABLE, A.CT_E.SERVER_IS_NOT_ACCESSABLE),
                )
        #
        self.disk_resource_status_map = {}
        for index in (0, 1):
            computer_list: list[ComputerDescription] = A.D.map(
                lambda item: item[0],
                A.D.filter(
                    lambda item: item[1] == index,
                    RESOURCE_LIST.DISK_STATISTICS_COMPUTER_DESCRIPTION,
                ),
            )  # type: ignore
            with PIHThreadPoolExecutor(max_workers=len(computer_list)) as executor:
                for disk_statistics_computer_description in computer_list:
                    executor.submit(
                        lambda item: self.set_disk_resource_status(item, index == 1),
                        disk_statistics_computer_description,
                    )
        #
        print("Время:", str(time.perf_counter() - start))
        print("Resources", self.resources_status_map)
        print("Disk statistics", self.disk_resource_status_map)



import ipih

from pih import A
from pih.consts.hosts import Hosts
from pih.collections.service import ServiceDescription
from pih.collections import ResourceStatus, ComputerDescription, SiteResourceStatus

NAME: str = "Checker"

HOST = Hosts.BACKUP_WORKER.NAME

VERSION: str = "0.13"

SD: ServiceDescription = ServiceDescription(
    name=NAME,
    description="Checker service",
    host=HOST,
    commands=("get_resource_status_list",),
    standalone_name="chk",
    use_standalone=True,
    version=VERSION,
)


class RESOURCE_LIST:

    INTERNET: ResourceStatus = A.D.fill_data_from_source(
        ResourceStatus(), A.CT_R_D.INTERNET
    )

    VPN_PACS_SPB: ResourceStatus = A.D.fill_data_from_source(
        ResourceStatus(), A.CT_R_D.VPN_PACS_SPB
    )

    PACS_SPB: ResourceStatus = A.D.fill_data_from_source(
        ResourceStatus(), A.CT_R_D.PACS_SPB
    )

    SITE: list[SiteResourceStatus] = []

    WAPPI: list[ResourceStatus] = []

    WS: list[ResourceStatus] = []

    SERVER: list[ResourceStatus] = []

    POLIBASE1: ResourceStatus = A.D.fill_data_from_source(
        ResourceStatus(), A.CT_R_D.POLIBASE1
    )

    POLIBASE2: ResourceStatus = A.D.fill_data_from_source(
        ResourceStatus(), A.CT_R_D.POLIBASE2
    )

    DISK_STATISTICS_COMPUTER_DESCRIPTION: list[tuple[ComputerDescription, int]] = []



import ipih

from pih import A
from CheckerService.const import SD
from CheckerService.api import CheckerApi as Api
from pih.tools import e, ne, n, nn, BitMask as BM

SC = A.CT_SC

ISOLATED: bool = False

api: Api = Api()

from typing import Any
from datetime import datetime
from pih.tools import ParameterList
from pih.consts import CheckableSections
from pih.collections import (
    EventDS,
    Workstation,
    ResourceStatus,
    WSResourceStatus,
    ChillerIndicationsValueContainer,
)
from pih.collections.service import SubscribtionResult

CHILLER_ALERT_COUNT_DEFAULT: int = A.S.get(A.CT_S.CHILLER_COUNT_DEFAULT)


class DH:
    current_datetime: datetime | None = None
    action_at_work: bool = False
    chiller_alert_count: int = CHILLER_ALERT_COUNT_DEFAULT


def check_resources_handler(force_update: bool, current_datetime: datetime) -> None:
    DH.action_at_work = True
    api.check_resources(force_update, current_datetime)
    DH.action_at_work = False


def service_call_handler(sc: SC, parameter_list: ParameterList) -> Any:
    if sc == SC.heart_beat:
        current_datetime: datetime = A.D_Ex.parameter_list(parameter_list).get()
        DH.current_datetime = current_datetime
        if not DH.action_at_work:
            check_resources_handler(False, current_datetime)
        return True
    if sc == SC.register_chiller_indications_value:
        subscribtion_result: SubscribtionResult | None = A.D_Ex.subscribtion_result(
            parameter_list
        )
        if nn(subscribtion_result):
            result: dict[str, Any] | None = subscribtion_result.result
            if ne(result):
                event_type: A.CT_E | None = None
                indications_value_container: ChillerIndicationsValueContainer = (
                    A.D.fill_data_from_source(
                        ChillerIndicationsValueContainer(), result
                    )
                )
                if (
                    indications_value_container.indicators
                    == A.CT_I.CHILLER.INDICATOR_EMPTY_DISPLAY
                ):
                    event_type = A.E_B.chiller_was_turned_off()
                    if A.C_IND.chiller_on():
                        A.E.send(
                            event_type,
                        )
                    else:
                        if DH.current_datetime.minute == 0:
                            A.E.send(
                                event_type,
                                flags=BM.remove(
                                    BM.set(A.D.get(event_type).flags),
                                    A.CT_L_ME_F.SAVE,
                                ),
                            )
                else:
                    if nn(indications_value_container.temperature):
                        event_off: EventDS | None = A.R.get_first_item(
                            A.R_E.get_last(A.E_B.chiller_was_turned_off())
                        )
                        event_type_on: A.CT_E = A.E_B.chiller_was_turned_on()
                        if nn(event_off):
                            event_on: EventDS | None = A.R.get_first_item(
                                A.R_E.get_last(event_type_on)
                            )
                            if n(event_on) or not A.C_IND.chiller_on():
                                A.E.send(event_type_on)
                        if indications_value_container.temperature > A.S.get(
                            A.CT_S.CHILLER_ALERT_TEMPERATURE
                        ):
                            DH.chiller_alert_count -= 1
                            if DH.chiller_alert_count <= 0:
                                event_type = A.E_B.chiller_temperature_alert_was_fired()
                                if A.C_E.timeouted(event_type, None, 5 * 60):
                                    A.E.send(event_type)
                        else:
                            if DH.chiller_alert_count < 0:
                                DH.chiller_alert_count = CHILLER_ALERT_COUNT_DEFAULT
                                event_type = (
                                    A.E_B.chiller_temperature_alert_was_resolved()
                                )
                                if A.C_E.timeouted(
                                    event_type, timeout_in_seconds=5 * 60
                                ):
                                    A.E.send(event_type)
                return True
        return False
    if sc == SC.get_resource_status_list:
        section_list: list[str] | None = parameter_list.next()
        force_update: bool = parameter_list.next()
        if force_update:
            api.check_resources(True)
        result: list[ResourceStatus] = []
        if e(section_list):
            result = A.D.to_list(api.resources_status_map)
        else:
            all: bool = parameter_list.next()
            section: str | None = None
            for section in section_list:
                section = section.upper()
                if section == CheckableSections.RESOURCES.name:
                    result += A.D.to_list(api.resources_status_map)
                if section == CheckableSections.WS.name:
                    if all:

                        def map_function(
                            workstation: Workstation,
                        ) -> WSResourceStatus:
                            return WSResourceStatus(
                                workstation.name,
                                f"Компьютер: {workstation.name} ({workstation.description})",
                                None,
                                workstation.accessable,
                                None,
                            )

                        result += A.R.map(map_function, A.R_WS.all()).data
                    else:
                        result += A.D.to_list(api.ws_status_map)
                if section == CheckableSections.SERVERS.name:
                    result += A.D.to_list(api.server_status_map)
                if section == CheckableSections.DISKS.name:
                    result += A.D.to_list(api.disk_resource_status_map)

        return A.R.pack(A.CT_FCA.VALUE, result)
    return True


def service_starts_handler() -> None:
    A.SRV_A.subscribe_on(SC.heart_beat)
    A.SRV_A.subscribe_on(SC.register_chiller_indications_value)


def start(as_standalone: bool = False) -> None:
    A.SRV_A.serve(
        SD,
        service_call_handler,
        service_starts_handler,
        isolate=ISOLATED,
        as_standalone=as_standalone,
    )


if __name__ == "__main__":
    start()
